#include <linux/fs.h>       /* inode, file, file_operations, ... */
#include <linux/kernel.h>   /* printk() */
#include <linux/module.h>
#include <linux/sched.h>    /* task_struct */
#include <linux/timer.h> 


MODULE_AUTHOR( "Ryo" );
MODULE_DESCRIPTION( "Sample character device driver." );
MODULE_LICENSE( "GPL" );


static int   devmajor = 77;
static char* devname  = "chardev";
static char* msg      = "module [timer.o]";


/* 500 tick = 5 sec. */
static int timeout = 500; 
MODULE_PARM( timeout, "i" );

static struct timer_list chardev_timer;
static int timerstop = 0;

static void chardev_timer_handler( unsigned long data );
static void chardev_del_timer();


/*
 * ϥեޥåѴؿ
 */
static void 
log_format( char* body )
{
	if ( current != NULL ) {
		printk( KERN_INFO "%s : PID = %-5i cmdline = \"%-10s\" : %s\n", 
			msg, current->pid, current->comm, body );
	}
	else {
		printk( KERN_INFO "%s : %s\n", msg, body );
	}
}


/* 
 * ޡϿ 
 */
static void 
chardev_timer_register( void )
{
	init_timer( &chardev_timer );

	chardev_timer.expires  = jiffies + timeout;
	chardev_timer.data     = ( unsigned long )jiffies;
	chardev_timer.function = chardev_timer_handler;

	printk( "%s : Register timer (%ld)\n", msg, jiffies );
	add_timer( &chardev_timer );
}


/* 
 * ॢȽ 
 */
static void 
chardev_timer_handler( unsigned long data )
{
	printk( KERN_INFO "%s : Timeout ( registered %ld, now %ld )\n", 
		msg, data, jiffies );

	if ( timerstop ) {
		printk( KERN_INFO "%s : Timer is not to be restarted.\n", msg );
		return;
	}

	chardev_timer_register();
}


/*
 * open()
 */
static int 
chardev_open( struct inode* inode, struct file* filp )
{
	log_format( "open()  called" );
	return 0;
}


/*
 * release()
 */
static int 
chardev_release( struct inode* inode, struct file* filp )
{
	log_format( "close() called" );
	return 0;
}


/*
 * read()
 * ޡؿ¹Ԥ롣
 */
static ssize_t 
chardev_read( struct file* filp, char* buf, size_t count, loff_t* pos )
{
	log_format( "read()  called" );
	log_format( "Delete registered timer." );

	chardev_del_timer();
	return 0;
}


/*
 * ޡؿ
 * ߥե饰ꤷdel_timer_sync Ǵ¸Υޡؿ
 * ʼ¹ξϽλޤԤġˡ
 */
static void
chardev_del_timer()
{
	timerstop = 1;
	del_timer_sync( &chardev_timer );
}


static struct file_operations chardev_fops = 
{
	owner   : THIS_MODULE,
	read    : chardev_read,
	open    : chardev_open,
	release : chardev_release,
};


/*
 * ⥸塼ν
 * insmod ˸ƤФ
 */
int 
init_module( void )
{
	if ( register_chrdev( devmajor, devname, &chardev_fops ) ) {
		log_format( "register_chrdev failed" );
		return -EBUSY;
	}

	chardev_timer_register();

	log_format( "loaded  into kernel" );
	return 0;
}


/*
 * ⥸塼β
 * ⥸塼λȿ 0 ǤСrmmod ˸ƤФ
 */
void 
cleanup_module( void )
{
	chardev_del_timer();

	if ( unregister_chrdev( devmajor, devname ) ) {
		log_format( "unregister_chrdev failed" );
		/* к̵ ... */
	}

	log_format( "removed from kernel" );
}


/* End of timer.c */
