#include <asm/io.h>          /* outb(), inb() */
#include <linux/interrupt.h> /* tasklet */
#include <linux/ioport.h>    /* check_region(), request_region(), release_region() */
#include <linux/kernel.h>    /* printk() */
#include <linux/module.h>
#include <linux/sched.h>     /* task_struct */
#include <linux/spinlock.h>


MODULE_LICENSE( "GPL" );


#define IOPORT_BASE     0x378
#define IOPORT_RANGE    3

#define IOPORT_DATA     ( IOPORT_BASE + 0 )
#define IOPORT_CONTROL  ( IOPORT_BASE + 2 )

#define IRQ_NUMBER      7

#define DATA_LED_3      2
#define DATA_PULLUP     3
#define IRQ_ENABLE      4


static char* devname  = "tasklet";
static char* msg      = "module [tasklet.o]";

static unsigned int initial_data;
static unsigned int initial_control;

static void tasklet_body( unsigned long value );
DECLARE_TASKLET( tasklet_fire_led, tasklet_body, ( unsigned long )NULL );

static spinlock_t tasklet_spinlock;


/*
 * ߥϥɥ
 */
static void
tasklet_interrupt( int irq, void* dev_id, struct pt_regs* regs )
{
	tasklet_schedule( &tasklet_fire_led );
}


/*
 * LED ѥåȼ
 */
void 
tasklet_body( unsigned long value )
{
	unsigned long flags;
	unsigned int  data;

	printk( KERN_INFO "%s : Tasklet fire-led being executed.\n", msg );

	spin_lock_irqsave( &tasklet_spinlock, flags );

	data  = inb( IOPORT_DATA );
	data ^= ( 1 << DATA_LED_3 );
	outb( data, IOPORT_DATA );

	spin_unlock_irqrestore( &tasklet_spinlock, flags );
}


/*
 * ⥸塼ν
 * insmod ˸ƤФ
 */
int 
init_module( void )
{
	unsigned int data;

	if ( request_region( IOPORT_BASE, IOPORT_RANGE, devname ) == NULL ) {
		printk( KERN_INFO "%s : request_region returned NULL\n", msg );
		return -EBUSY;
	}


	if ( request_irq( IRQ_NUMBER, tasklet_interrupt, 0, devname, NULL ) ) {
		printk( KERN_INFO "%s : request_irq failed\n", msg );
		release_region( IOPORT_BASE, IOPORT_RANGE );
		return -EBUSY;
	}

	spin_lock_init( &tasklet_spinlock );

	initial_data    = inb( IOPORT_DATA );
	initial_control = inb( IOPORT_CONTROL );
	rmb();

	data = inb( IOPORT_DATA );
	data |= ( 1 << DATA_PULLUP );
	outb( data, IOPORT_DATA );
	mb();

	data = inb( IOPORT_CONTROL );
	data |= ( 1 << IRQ_ENABLE );
	outb( data, IOPORT_CONTROL );
	mb();
	
	printk( KERN_INFO "%s : loaded  into kernel\n", msg );
	return 0;
}


/*
 * ⥸塼β
 * ⥸塼λȿ 0 ǤСrmmod ˸ƤФ
 */
void 
cleanup_module( void )
{
	free_irq( IRQ_NUMBER, NULL );
	tasklet_kill( &tasklet_fire_led );

	outb( initial_control, IOPORT_CONTROL );
	outb( initial_data,    IOPORT_DATA    );

	release_region( IOPORT_BASE, IOPORT_RANGE );

	printk( KERN_INFO "%s : removed from kernel\n", msg );
}


/* End of tasklet.c */
