#include <asm/types.h>    /* u8, ... */
#include <linux/errno.h>
#include <linux/ioport.h> /* IORESOURCE_IO, ... */
#include <linux/module.h> /* module ˤɬ */
#include <linux/kernel.h> /* printk */
#include <linux/pci.h>


MODULE_LICENSE( "GPL" );

static char* msg = "module [pci.o]";


/*
 * PCI Х³ƤǥХξɽ롣
 */
static void 
print_pci_information( void )
{
	struct pci_dev* device = NULL;

	while ( 1 ) {
		/*
		 * vendor id  device id  PCI_ANY_ID ꤹ뤳Ȥ
		 * ³ƤƤ PCI ǥХоݤȤƤ롣
		 */
		device = pci_find_device( PCI_ANY_ID, PCI_ANY_ID, device );

		if ( device == NULL )
			break;


		{
			#define BASE_ADDRESS_NUM 6

			u16 vendor_id;
			u16 device_id;
			u16 class;
			u16 sub_vendor_id;
			u16 sub_device_id;
			u8  irq;

			unsigned long resource_start;
			unsigned long resource_end;
			unsigned long resource_flag;

			int i;


			pci_read_config_word( device, PCI_VENDOR_ID,           &vendor_id );
			pci_read_config_word( device, PCI_DEVICE_ID,           &device_id );
			pci_read_config_word( device, PCI_CLASS_DEVICE,        &class );
			pci_read_config_word( device, PCI_SUBSYSTEM_VENDOR_ID, &sub_vendor_id );
			pci_read_config_word( device, PCI_SUBSYSTEM_ID,        &sub_device_id );
			pci_read_config_byte( device, PCI_INTERRUPT_LINE,      &irq );


			printk( KERN_INFO "%s : vendor id = %i, device id = %i\n",
				msg, vendor_id, device_id );

			printk( KERN_INFO "%s : class = %i\n", msg, class );
			printk( KERN_INFO "%s : subsystem\n", msg );

			printk( KERN_INFO "%s :   vendor id = %i, device id = %i\n",
				msg, sub_vendor_id, sub_device_id );

			printk( KERN_INFO "%s : IRQ = %i\n", msg, irq );


			for ( i = 0; i < BASE_ADDRESS_NUM; i ++ ) {
				resource_start = pci_resource_start( device, i );
				resource_end   = pci_resource_end( device, i );
				resource_flag  = pci_resource_flags( device, i );

				if ( resource_start != 0 || resource_end != 0 ) {
					printk( KERN_INFO "%s : Base address 0x%0x - 0x%0x\n", 
						msg, resource_start, resource_end );

					if ( resource_flag & IORESOURCE_IO )
						printk( KERN_INFO "%s :   I/O port\n", msg );

					if ( resource_flag & IORESOURCE_MEM )
						printk( KERN_INFO "%s :   Memory\n", msg );

					if ( resource_flag & IORESOURCE_PREFETCH )
						printk( KERN_INFO "%s :   Prefetchable\n", msg );

					if ( resource_flag & IORESOURCE_READONLY )
						printk( KERN_INFO "%s :   Read Only\n", msg );

					printk( KERN_INFO "%s : \n", msg );
				}
			}

			printk( KERN_INFO "%s : \n", msg );
		}
	}
}


/*
 * ⥸塼ν
 * insmod ˸ƤФ
 */
int 
init_module( void )
{
    printk( KERN_INFO "%s : loaded  into kernel\n", msg );

    print_pci_information();

    /* Ȥ߹ޤ̵̣Τǡ˥顼֤Ȥ
       Ȥ߹ޤʤͤˤƤ */

    return -ENODEV;
}


/*
 * ⥸塼β
 * ⥸塼λȿ 0 ǤСrmmod ˸ƤФ
 */
void 
cleanup_module( void )
{
    printk( KERN_INFO "%s : removed from kernel\n", msg );
}


/* End of pci.c */
