#include <asm/uaccess.h>    /* copy_from_user */
#include <linux/errno.h>
#include <linux/module.h>   /* module ˤɬ */
#include <linux/kernel.h>   /* printk */
#include <linux/proc_fs.h>
#include <linux/stat.h>     /* S_IFREG, ... */


MODULE_LICENSE( "GPL" );

static char* msg = "module [proc.o]";

#define MAXBUF 32
static unsigned char procbuf[ MAXBUF ];
static int buf_pos;

static struct semaphore proc_sem;


/*
 * /proc/driver/sampleproc Фɤ߽ФԤäݤ˼¹Ԥ롣
 * ХåեƤ֤
 */
static int
proc_read( char* page, char** start, off_t offset, int count, int* eof, void* data )
{
	int len = 0;

	printk( KERN_INFO "%s : read()  called\n", msg );

	if ( down_interruptible( &proc_sem ) ) {
		printk( KERN_INFO "%s : down_interruptible for read failed\n", msg );
		return -ERESTARTSYS;
	}

	if ( buf_pos != 0 ) {
		len += sprintf( page, "%s : Internal buffer contents = %s\n", msg, procbuf );
		buf_pos = 0;
	}

	up( &proc_sem );

	*eof = 1;
	return len;
}


/**
 * /proc/driver/sampleproc Фƽ񤭹ߤԤäݤ˼¹Ԥ롣
 * 񤭹ޤ줿ƤХåեݻ롣
 */
static int
proc_write( struct file* filp, const char* buffer, unsigned long count, void* data )
{
	int copy_len;

	printk( KERN_INFO "%s : write() called\n", msg );

	if ( down_interruptible( &proc_sem ) ) {
		printk( KERN_INFO "%s : down_interruptible for write failed\n", msg );
		return -ERESTARTSYS;
	}

	if ( buf_pos == MAXBUF ) {
		up( &proc_sem );
		printk( KERN_INFO "%s : no space left\n", msg );
		return -ENOSPC;
	}

	if ( count > ( MAXBUF - buf_pos ) )
		copy_len = MAXBUF - buf_pos;
	else
		copy_len = count;

	if ( copy_from_user( procbuf + buf_pos, buffer, copy_len ) ) {
		up( &proc_sem );
		printk( KERN_INFO "%s : copy_from_user failed\n", msg );
		return -EFAULT;
	}

	buf_pos += copy_len;
	up( &proc_sem );

	return copy_len;
}


/*
 * ⥸塼ν
 * insmod ˸ƤФ
 */
int 
init_module( void )
{
	struct proc_dir_entry* entry;
	entry = create_proc_entry( "driver/sampleproc", 
		       S_IFREG | S_IRUGO | S_IWUGO, NULL );

	if ( entry != NULL ) {
		entry->read_proc  = proc_read;
		entry->write_proc = proc_write;
	}
	else {
		printk( KERN_INFO "%s : create_proc_entry failed\n", msg );
		return -EBUSY;
	}

	sema_init( &proc_sem, 1 );
	printk( KERN_INFO "%s : loaded  into kernel\n", msg );

	return 0;
}


/*
 * ⥸塼β
 * ⥸塼λȿ 0 ǤСrmmod ˸ƤФ
 */
void 
cleanup_module( void )
{
    remove_proc_entry( "driver/sampleproc", NULL );
    printk( KERN_INFO "%s : removed from kernel\n", msg );
}


/* End of proc.c */
